/**
  *  \file data.cpp
  *  \brief Integrator Data Management
  */

#include <map>
#include <ctype.h>
#include "data.h"
#include "util.h"
#include "game.h"
#include "race.h"
#include "enumerator.h"
#include "program.h"
#include "action.h"
#include "nls.h"
#include "osdep.h"
#include "loaderstate.h"
#include "net/config.h"

/*************************** Data::ReaderState ***************************/

/** Reader state. Offers a basic line parsing primitive, and keeps
    track of the current line number. */
class Data::ReaderState {
    std::istream& in;

    int lineno;

 public:
    ReaderState(std::istream& in)
        : in(in), lineno(0)
        { }

    Line readLine(std::string& key, std::string& value);

    void fail(std::string s);
};

/** Read a line.
    \param key   [out] Key of a key/value line
    \param value [out] Value of a key/value line
    \return line type */
Data::Line
Data::ReaderState::readLine(std::string& key, std::string& value)
{
    key.erase();
    value.erase();
    while (1) {
        std::string the_line;
        if (!getline(in, the_line))
            return EndOfFile;

        ++lineno;

        // trim trailing CR
        if (the_line.size() != 0 && the_line[the_line.size()-1] == '\r')
            the_line.erase(the_line.size()-1);

        // trim space
        the_line = strTrim(the_line);

        // empty line? comment? Try again
        if (the_line.size() == 0 || the_line[0] == '#' || the_line[0] == ';')
            continue;

        // end?
        if (the_line == "}")
            return ClosingBrace;

        // ends with "{"?
        bool is_open = false;
        if (the_line[the_line.size()-1] == '{') {
            is_open = true;
            the_line.erase(the_line.size()-1);
            the_line = strRTrim(the_line);
        }

        // split into halves
        std::string::size_type i = the_line.find_first_of(" \t");
        if (i == the_line.npos)
            fail(_("Syntax error: missing space after keyword name"));

        // figure out values
        key.assign(the_line, 0, i);
        value = strTrim(the_line.substr(i+1));
        if (value.size() && value[0] == '"') {
            value.erase(0, 1);
            if (value.size() && value[value.size()-1] == '"')
                value.erase(value.size()-1);
            std::string::size_type i = 0;
            while (i < value.size()) {
                if (value[i] == '\\')
                    value.erase(i, 1);
                ++i;
            }
        }
        return is_open ? Subtree : Assignment;
    }
}

/** Error message. Throws an exception containing the specified
    message and the current line number. */
void
Data::ReaderState::fail(std::string s)
{
    ::fail(_("Line ") + itoa(lineno) + ": " + s);
}

/*************************** Data::WriterState ***************************/

/** Writer state. Keeps track of current indent, and is used as a
    place-holder for various enumerator functions. */
class Data::WriterState : public KVEnumerator,
                          public ValueEnumerator<Race>,
                          public ValueEnumerator<Action>
{
    std::ostream& out;

    int indent;

    Data& data;

 public:
    WriterState(std::ostream& out, Data& data)
        : out(out), indent(0), data(data)
        { }

    void writeLine(std::string key, std::string value, bool subtree = false);

    void writeClosingBrace();

    // KVEnumerator:
    void handleItem(std::string key, std::string, std::string value);

    // ValueEnumerator<T>:
    void handleItem(Race& r);
    void handleItem(Action& r);
};

/** Write a key/value line.
    \param key     [in] Key
    \param value   [in] Value
    \param subtree [in] If set, make a new subtree. Otherwise, just make a simple
                   key/value line. */
void
Data::WriterState::writeLine(std::string key, std::string value, bool subtree)
{
    std::string::size_type i;
    out << std::string(indent, ' ') << key << ' ';
    if ((i = value.find_first_of("\"\\{")) != value.npos) {
        // we quote quotes, backslashes, and braces. The latter are
        // only special to avoid a line being interpreted as a group
        // opener, but quoting doesn't hurt.
        do {
            value.insert(i, "\\");
        } while ((i = value.find_first_of("\"\\{", i+2)) != value.npos);
        out << "\"" << value << "\"";
    } else {
        out << value;
    }
    if (subtree) {
        out << " {\n";
        indent += 4;
    } else {
        out << "\n";
    }
}

/** Write the closing brace that ends a subtree. */
void
Data::WriterState::writeClosingBrace()
{
    if (indent > 4)
        indent -= 4;
    else
        indent = 0;
    out << std::string(indent, ' ') << "}\n";
}

/** KVEnumerator: write key/value line. */
void
Data::WriterState::handleItem(std::string key, std::string, std::string value)
{
    if (value.size() && key != "-")
        writeLine(key, value);
}

/** ValueEnumerator<Race>: write a race definition. */
void
Data::WriterState::handleItem(Race& r)
{
    data.writeRace(r, *this);
}

/** ValueEnumerator<Action>: write a action definition. */
void
Data::WriterState::handleItem(Action& r)
{
    data.writeAction(r, *this);
}

/********************************** Data *********************************/

/** Constructor. */
Data::Data()
{ }

/** Destructor. */
Data::~Data()
{
    safeClear(m_Games);
    safeClear(m_Programs);
}

/** Read a subtree into a PropertyObject.
    \param p   [in/out] The object to populate
    \param st  [in] Reader state
    \return true if we end with a ClosingBrace, false if we end with EOF */
bool
Data::readAny(PropertyObject& p, ReaderState& rst, LoaderState& st)
{
    std::string key, value;
    while (1) {
        switch (rst.readLine(key, value)) {
         case Subtree:
            if (PropertyObject* pp = p.createSubtree(key, value, st)) {
                if (!readAny(*pp, rst, st))
                    rst.fail(_("Unexpected end of file"));
            } else {
                readUnknown(rst);
            }
            break;
         case EndOfFile:
            return false;
         case ClosingBrace:
            return true;
         case Assignment:
            p.assign(key, value, st);
            break;
        }
    }
}

/** Read unknown subtree. Here, we don't have a PropertyObject as in
    readAny(), so we discard everything we get.  */
void
Data::readUnknown(ReaderState& rst)
{
    std::string key, value;
    int level = 1;
    while (1) {
        switch (rst.readLine(key, value)) {
         case Subtree:
            ++level;
            break;
         case ClosingBrace:
            if (--level == 0)
                return;
            break;
         case Assignment:
            break;
         case EndOfFile:
            rst.fail(_("Unexpected end of file"));
        }
    }
}

/** Read everything, main entry point. */
void
Data::read(std::istream& in, LoaderState& st)
{
    ReaderState rst(in);
    st.start();
    if (readAny(*this, rst, st))
        rst.fail(_("Unexpected '}' in input"));
}

/** Write a game definition. */
void
Data::writeGame(Game& g, WriterState& st)
{
    st.writeLine("game", g.getName(), true);
    g.enumerate(st);
    g.enumerateRaces(st);
    st.writeClosingBrace();
}

/** Write a race definition. */
void
Data::writeRace(Race& r, WriterState& st)
{
    st.writeLine("race", itoa(r.getRaceNumber()), true);
    r.enumerate(st);
    st.writeClosingBrace();
}

/** Write a program definition. */
void
Data::writeProgram(Program& p, WriterState& st)
{
    st.writeLine("program", p.getName(), true);
    p.enumerate(st);
    p.enumerateActions(st);
    st.writeClosingBrace();
}

/** Write a action definition. */
void
Data::writeAction(Action& a, WriterState& st)
{
    st.writeLine("action", a.getName(), true);
    a.enumerate(st);
    st.writeClosingBrace();
}

/** Writing, main entry point. */
void
Data::write(std::ostream& out)
{
    WriterState st(out, *this);

    out << _("#\n#  Game List\n#\n");
    for (std::vector<Game*>::iterator i = m_Games.begin(); i != m_Games.end(); ++i)
        writeGame(**i, st);

    out << _("\n\n#\n#  Program List\n#\n");
    for (std::vector<Program*>::iterator i = m_Programs.begin(); i != m_Programs.end(); ++i)
        writeProgram(**i, st);

    out << _("\n\n#\n#  Configuration\n#\n");
    st.writeLine("config", "http", true);
    HttpConfig::instance.enumerate(st);
    st.writeClosingBrace();
    st.writeLine("config", "smtp", true);
    SmtpConfig::instance.enumerate(st);
    st.writeClosingBrace();
}

/** Add a new game. */
void
Data::addGame(Game* g)
{
    safeAppend(m_Games, g);
}

/** Add a new program. */
void
Data::addProgram(Program* p)
{
    safeAppend(m_Programs, p);
}

/** Enumerate all games. */
void
Data::enumerateGames(ValueEnumerator<Game>& e)
{
    enumerateVector(m_Games, e);
}

/** Enumerate all programs. */
void
Data::enumeratePrograms(ValueEnumerator<Program>& e)
{
    enumerateVector(m_Programs, e);
}

/** Delete a game. */
void
Data::deleteGame(Game* g)
{
    for (std::vector<Game*>::iterator i = m_Games.begin(); i != m_Games.end(); ++i) {
        if (*i == g) {
            m_Games.erase(i);
            break;
        }
    }
    delete g;
}

/** Find a game by name.
    \return game if found, 0 otherwise */
Game*
Data::getGameByName(std::string name)
{
    for (std::vector<Game*>::iterator i = m_Games.begin(); i != m_Games.end(); ++i)
        if ((**i).getName() == name)
            return *i;
    return 0;
}

/** Find a program by name.
    \return program if found, 0 otherwise */
Program*
Data::getProgramByName(std::string prog)
{
    for (std::vector<Program*>::iterator i = m_Programs.begin(); i != m_Programs.end(); ++i)
        if ((**i).getName() == prog)
            return *i;
    return 0;
}

/** Find program that has a feature.
    \param feature      Feature to find
    \param invert       If specified and true, look for programs not having that feature
    \return first found program, or 0 */
Program*
Data::getProgramByFeature(std::string feature, bool invert)
{
    for (std::vector<Program*>::iterator i = m_Programs.begin(); i != m_Programs.end(); ++i)
        if ((**i).hasFeature(feature) != invert)
            return *i;
    return 0;
}

void
Data::enumerate(KVEnumerator& e)
{
    e.handleItem("basedir", _("Base directory for games"), m_GameBaseDir);
}

bool
Data::assign(std::string name, std::string value, LoaderState& st)
{
    if (name == "basedir") {
        if (st.isAccept(LoaderState::Globals))
            m_GameBaseDir = value;
        return true;
    }
    return false;
}

PropertyObject*
Data::createSubtree(std::string key, std::string value, LoaderState& st)
{
    if (key == "game") {
        if (st.isAccept(LoaderState::Games)) {
            Game* g = getGameByName(value);
            if (!g) {
                g = new Game(value);
                addGame(g);
            }
            return g;
        } else {
            return 0;
        }
    }
    if (key == "program") {
        if (st.isAccept(LoaderState::Programs)) {
            Program* p = getProgramByName(value);
            if (!p) {
                p = new Program(value);
                addProgram(p);
            }
            return p;
        } else {
            return 0;
        }
    }
    if (key == "config") {
        if (st.isAccept(LoaderState::Globals)) {
            if (value == "http")
                return &HttpConfig::instance;
            else if (value == "smtp")
                return &SmtpConfig::instance;
            else
                return 0;
        } else {
            return 0;
        }
    }
    return 0;
}

namespace {
    struct ActionCollector : public ValueEnumerator<Action> {
        std::map<std::string,Action*> m_Stuff;

        void handleItem(Action& a);
    };

    void ActionCollector::handleItem(Action& a) {
        std::map<std::string,Action*>::iterator i = m_Stuff.find(a.getName());

        if (i == m_Stuff.end() || i->second->getPreference() < a.getPreference())
            m_Stuff[a.getName()] = &a;
    }
}

/** Enumerate all simple actions for a game. */
void
Data::enumerateSimpleActions(Game& /*g*/, ValueEnumerator<Action>& a)
{
    ActionCollector coll;

    for (std::vector<Program*>::iterator i = m_Programs.begin(); i != m_Programs.end(); ++i) {
        // FIXME: would check whether this is applicable to this game
        (**i).enumerateActions(coll);
    }

    for (std::map<std::string,Action*>::iterator i = coll.m_Stuff.begin(); i != coll.m_Stuff.end(); ++i)
        a.handleItem(*i->second);
}

/** Get local directory name for a new game. If user uses Winplan and all
    slots are used up, returns empty string. Specify force=true to make
    this create a directory outside the Winplan tree.
    \param gamename Name of game
    \param force Force generation of a name */
std::string
Data::getNewGameDirectory(std::string gamename, bool force)
{
    // If we have Winplan, try a Winplan game slot
    if (Program* p = getProgramByFeature("winplan")) {
        std::string progdir = p->getDirectory();
        for (int i = 1; i <= 8; ++i) {
            char w[10];
            std::sprintf(w, "vpwork%d", i);
            std::string mydir = makeFileName(progdir, w);
            bool ok = true;
            for (std::vector<Game*>::iterator g = m_Games.begin(); g != m_Games.end(); ++g)
                if (strLower(mydir) == strLower((**g).getLocalDirectory()))
                    ok = false;
            if (ok)
                return mydir;
        }
        if (!force)
            return std::string();
    }

    // Make my own directory
    std::string basename = strTrim(std::string(gamename, 0, gamename.find('@')));
    if (basename.empty())
        basename = gamename;

    // Strip all funny characters
    std::string::size_type i = 0;
    while (i < basename.size()) {
        if (!isalnum((unsigned char) basename[i]) && basename[i] != '_' && basename[i] != '-')
            basename.erase(i, 1);
        else
            ++i;
    }

    if (basename.empty())
        basename = "game";

    // Trim to 8 characters, due to DOS file names
    bool need_short_name = (getProgramByFeature("lfn", true) != 0);
    if (need_short_name && basename.size() > 8) {
        // if it's a game like "BackToBasic17", we'll reduce
        // everything before the digits. Otherwise, we'll simply
        // truncate.
        std::string::size_type n = basename.size();
        std::string::size_type num_digits = 0;
        while (n && basename[n-1] >= '0' && basename[n-1] <= '9')
            --n, ++num_digits;

        if (num_digits <= 5) {
            // not more than 5 digits
            basename.erase(8 - num_digits, basename.size() - 8);
        } else {
            // too many
            basename.erase(8);
        }
    }

    // Available? Make it unique if not
    int counter = 0;
    std::string result = makeFileName(m_GameBaseDir, basename);
    while (isExistingFileName(result)) {
        if (++counter > 10000)
            break;
        std::string suffix = itoa(counter);
        if (need_short_name && suffix.size() + basename.size() > 8)
            result = makeFileName(m_GameBaseDir, basename.substr(0, 8 - suffix.size()) + suffix);
        else
            result = makeFileName(m_GameBaseDir, basename + suffix);
    }
    return result;
}
