/**
  *  \file encoding.cpp
  *  \brief Various en- and decoders
  */

#include <cstring>
#include "encoding.h"

static const char base64_alphabet[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
static const char hex_alphabet[] = "0123456789ABCDEF";

static int
hexValue(char c)
{
    if (c >= '0' && c <= '9')
        return c - '0';
    if (c >= 'A' && c <= 'F')
        return c - 'A' + 10;
    if (c >= 'a' && c <= 'f')
        return c - 'a' + 10;
    return -1;
}

static int
b64Value(char c)
{
    const char* p = std::strchr(base64_alphabet, c);
    if (p != 0)
        return p - base64_alphabet;
    else
        return 0;
}

/** Encode Base64. */
std::string
encodeBase64(const std::string input)
{
    std::string result;
    std::size_t index = 0;
    result.reserve((input.size() + 2) / 3 * 4);
    while (index + 3 <= input.size()) {
        unsigned long value = 0;
        value += 65536*(unsigned char) input[index++];
        value += 256*(unsigned char) input[index++];
        value += (unsigned char) input[index++];
        result += base64_alphabet[(value >> 18) & 63];
        result += base64_alphabet[(value >> 12) & 63];
        result += base64_alphabet[(value >> 6) & 63];
        result += base64_alphabet[(value) & 63];
    }

    unsigned long value = 0;
    switch (input.size() - index) {
     case 2:
        value += 65536*(unsigned char) input[index++];
        value += 256*(unsigned char) input[index++];
        result += base64_alphabet[(value >> 18) & 63];
        result += base64_alphabet[(value >> 12) & 63];
        result += base64_alphabet[(value >> 6) & 63];
        result += '=';
        break;
     case 1:
        value += 65536*(unsigned char) input[index++];
        result += base64_alphabet[(value >> 18) & 63];
        result += base64_alphabet[(value >> 12) & 63];
        result += '=';
        result += '=';
        break;
     default:
        break;
    }
    return result;
}

/** Decode Base64. */
std::string
decodeBase64(const std::string input)
{
    std::string result;

    // Trim trailing equal signs. Normal base64 stuff has
    // lengths divisible by four, implementing it this way
    // has the added benefit of getting robustness for
    // truncated streams for free.
    std::string::size_type limit = input.size();
    while (limit > 0 && input[limit-1] == '=')
        --limit;

    result.reserve(limit * 3 / 4 + 1);

    std::string::size_type index = 0;
    while (index < limit) {
        unsigned long value = 0;
        int nch = 0;
        switch (limit - index) {
         default:
            value += b64Value(input[index+3]);
            ++nch;
         case 3:
            value += b64Value(input[index+2]) << 6;
            ++nch;
         case 2:
            value += b64Value(input[index+1]) << 12;
            ++nch;
         case 1:
            value += b64Value(input[index+0]) << 18;
        }
        if (nch > 0)
            result.append(1, char((value >> 16) & 255));
        if (nch > 1)
            result.append(1, char((value >> 8) & 255));
        if (nch > 2)
            result.append(1, char((value) & 255));

        index += 4;
    }
    return result;
}

/** Encode URL-Encoding (percent-escaping). */
std::string
encodeUrl(std::string input)
{
    for (std::string::size_type i = 0; i < input.size(); ++i) {
        unsigned char u = input[i];
        if (u >= 0x80 || u < 0x20 || u == '%' || u == '&' || u == '+') {
            char tmp[2];
            input[i] = '%';
            tmp[0] = hex_alphabet[(u >> 4) & 15];
            tmp[1] = hex_alphabet[u & 15];
            input.insert(i+1, tmp, 2);
            i += 2;
        } else if (u == ' ') {
            input[i] = '+';
        }
    }
    return input;
}

/** Decode URL-Encoding (percent-escaping). */
std::string
decodeUrl(std::string input)
{
    std::string::size_type n = 0;
    while ((n = input.find('+', n)) != input.npos) {
        input[n] = ' ';
        ++n;
    }
    n = 0;
    while ((n = input.find('%', n)) != input.npos && input.size() - n >= 3) {
        int a = hexValue(input[n+1]);
        int b = hexValue(input[n+2]);
        if (a >= 0 && b >= 0) {
            input.erase(n+1, 2);
            input[n] = char(16*a + b);
        }
        ++n;
    }
    return input;
}

/** Encode quoted-printable. */
std::string
encodeQP(std::string input)
{
    for (std::string::size_type i = 0; i < input.size(); ++i) {
        unsigned char u = input[i];
        if (u >= 0x80 || u < 0x20 || u == '=') {
            char tmp[2];
            input[i] = '=';
            tmp[0] = hex_alphabet[(u >> 4) & 15];
            tmp[1] = hex_alphabet[u & 15];
            input.insert(i+1, tmp, 2);
            i += 2;
        }
    }
    return input;
}

/** Decode quoted-printable. */
std::string
decodeQP(std::string input)
{
    std::string::size_type n = 0;
    while ((n = input.find('=', n)) != input.npos && input.size() - n >= 3) {
        int a = hexValue(input[n+1]);
        int b = hexValue(input[n+2]);
        if (a >= 0 && b >= 0) {
            input.erase(n+1, 2);
            input[n] = char(16*a + b);
        }
        ++n;
    }
    return input;
}

#ifdef TRANSCODER
#include <iostream>

struct Huhu {
    const char* name;
    std::string (*f)(std::string);
} data[] = {
    { "e64", encodeBase64 },
    { "d64", decodeBase64 },
    { "eurl", encodeUrl },
    { "durl", decodeUrl },
    { "eqp", encodeQP },
    { "dqp", decodeQP },
    { 0, 0 }
};

int main()
{
    std::string s;
    int index = 0;
    while (std::cout << data[index].name << "> ", getline(std::cin, s)) {
        int found = -1;
        for (int i = 0; data[i].name != 0; ++i)
            if (s == data[i].name)
                found = i;
        if (found < 0) {
            std::cout << data[index].name << "('" << s << "') = '" << data[index].f(s) << "'\n";
        } else {
            index = found;
        }
    }
    return 0;
}
#endif
