/**
  *  \file enumerator.h
  */
#ifndef ENUMERATOR_H_INCLUDED
#define ENUMERATOR_H_INCLUDED

#include <string>
#include <vector>

/** Key/Value Enumerator. Interface for routines to enumerate
    key/value pairs (with descriptions). */
class KVEnumerator {
 public:
    KVEnumerator();

    virtual ~KVEnumerator();

    /** Handle an item.
        \param name        Name of item (variable name)
        \param description User-visible description
        \param value       String */
    virtual void handleItem(const std::string name,
                            const std::string description,
                            const std::string value) = 0;
};

/** Value Enumerator. Interface for routines to enumerate simple lists
    of values of type T. */
template<typename T>
class ValueEnumerator {
 public:
    ValueEnumerator()
        { }

    virtual ~ValueEnumerator();

    /** Handle an item.
        \param t The item. */
    virtual void handleItem(T& t) = 0;
};

template<typename T>
ValueEnumerator<T>::~ValueEnumerator()
{ }



// Specialized enumerators

/** Enumerate all items from a vector into a ValueEnumerator.
    \param vec [in] Vector
    \param e Receives all vector elements */
template<typename T>
void
enumerateVector(const std::vector<T*> vec, ValueEnumerator<T>& e)
{
    for (typename std::vector<T*>::const_iterator i = vec.begin(); i != vec.end(); ++i)
        e.handleItem(**i);
}

/** ValueEnumerator collecting items in a vector. This gathers up all
    items for later reference. */
template<typename T>
class VectorValueEnumerator : public ValueEnumerator<T> {
    std::vector<T*>& vec;

 public:
    /** Constructor.
        \param vec Vector that receives the items */
    VectorValueEnumerator(std::vector<T*>& vec)
        : vec(vec)
        { }
    
    void handleItem(T& t);
};

template<typename T>
void
VectorValueEnumerator<T>::handleItem(T& t)
{
    vec.push_back(&t);
}

#endif
