/**
  *  \file http.cpp
  *  \brief HTTP Access
  *  \author Stefan Reuther
  *
  *  Brief this module provides routines to perform HTTP requests.
  *
  *  References:
  *  - RFC 1945 HTTP/1.0
  *  - RFC 2068 HTTP/1.1
  *
  *  \todo evaluate returned headers: returned file name for
  *  downloads, status and location for redirects, etc.
  */

#include "net/http.h"
#include "util.h"
#include "net/socketstream.h"
#include "net/mimebuilder.h"
#include "net/connection.h"
#include "nls.h"

/******************************** Network ********************************/

static Url the_proxy;
static const char user_agent[] = "VGAPIntegrator/0.5";

/** Set proxy. */
void
Http::setProxy(const Url& proxy)
{
    the_proxy = proxy;
}

const Url&
Http::getProxy()
{
    return the_proxy;
}

/** Perform a request.
    \param host    [in] URL to access (uses host, port)
    \param request [in] Full text of request ("GET ..." plus headers plus blank line)
    \param body    [in] If non-null, body of the request (for POST)
    \param out     Receives reply body

    Reply headers are accumulated in the Http object.

    \return true if request successfully performed. HTTP errors ("404
    not found") also return success (as the actual HTTP transaction
    went ok); use isSuccessfulReply() to check for those. */
bool
Http::doRequest(const Url& host, const std::string request, MimeBuilder* body, std::ostream& out)
{
    // pick a port number
    int port = host.getPort();
    if (port < 0)
        port = 80;

    // connect
    if (!connect(host.getHost(), port)) {
        return false;
    }

    // Send request
    if (!m_Socket.write(request.data(), request.size())) {
        close();
        return false;
    }
    if (body) {
        SocketStream ss(m_Socket);
        body->write(ss);
    }

    // clear headers, we'll fill it again
    m_Headers.clear();

    // Read first result line
    std::string reply;
    if (!readLine(reply)) {
        close();
        return false;
    }

    if (reply.size() > 5 && reply.compare(0, 5, "HTTP/", 5) == 0) {
        // HTTP 1.0+ reply
        std::string::size_type n = reply.find(' ');
        if (n == reply.npos) {
            setMessage(_("Malformed HTTP Reply"), reply);
            close();
            return false;
        }
        reply.erase(0, n);
        reply = strLTrim(reply);
        // save answer
        m_ReplyCode = atoi(reply.c_str());
        m_ReplyLine = reply;
        // read headers
        while (1) {
            std::string tmp;
            if (!readLine(tmp)) {
                close();
                return false;
            }
            if (tmp.size() == 0)
                break;
            m_Headers.push_back(tmp);
        }
    } else {
        // HTTP 0.9 reply
        /* Note that this may break if the original line did not have
           network format (CRLF). However, since HTTP/0.9 cannot
           report any content type, we hope that it is (if at all)
           only used for HTML files */
        out << reply << "\r\n";
    }

    // Await reply
    char buffer[4096];
    int got;
    while ((got = m_Socket.read(buffer, sizeof buffer)) > 0)
        out.rdbuf()->sputn(buffer, got);

    m_Socket.close();
    return true;
}

/** Download a file (GET request).
    \param file [in] URL to query
    \param out  Will receive the result of the request
    \return true on success */
bool
Http::downloadFile(const Url& file, std::ostream& out)
{
    const Url* where_to_connect;
    std::string request;

    // build the request
    if (the_proxy.isValid()) {
        request = "GET " + file.toString() + " HTTP/1.0\015\012"
            + "User-Agent: " + user_agent + "\015\012"
            + "\015\012";
        where_to_connect = &the_proxy;
    } else {
        request = "GET " + file.getPath() + " HTTP/1.0\015\012"
            + "Host: " + file.getHost() + "\015\012"
            + "User-Agent: " + user_agent + "\015\012"
            + "\015\012";
        where_to_connect = &file;
    }
    return doRequest(*where_to_connect, request, 0, out);
}

/** Download a file (POST request).
    \param file [in] URL to query
    \param body [in/out] Body of request. This routine will
                remove the headers from the MimeBuilder.
    \param out  Will receive the result of the request
    \return true on success */
bool
Http::post(const Url& file, MimeBuilder& body, std::ostream& out)
{
    const Url* where_to_connect;
    std::string request;

    std::string mimeheaders = body.removeInitialHeaders();

    // build the request
    if (the_proxy.isValid()) {
        request = "POST " + file.toString() + " HTTP/1.0\015\012"
            + "User-Agent: " + user_agent + "\015\012"
            + "Content-Length: " + itoa(body.computeSize()) + "\015\012"
            + mimeheaders;
        where_to_connect = &the_proxy;
    } else {
        request = "POST " + file.getPath() + " HTTP/1.0\015\012"
            + "Host: " + file.getHost() + "\015\012"
            + "User-Agent: " + user_agent + "\015\012"
            + "Content-Length: " + itoa(body.computeSize()) + "\015\012"
            + mimeheaders;
        where_to_connect = &file;
    }

    return doRequest(*where_to_connect, request, &body, out);
}

/** Enumerate all headers. e.handleItem() will be called once for each
    header ("Content-Type: text/plain"). */
void
Http::enumerateHeaders(ValueEnumerator<std::string>& e)
{
    for (std::list<std::string>::iterator i = m_Headers.begin(); i != m_Headers.end(); ++i)
        e.handleItem(*i);
}

std::string
Http::getHeader(std::string name)
{
    name += ':';
    for (std::list<std::string>::iterator i = m_Headers.begin(); i != m_Headers.end(); ++i)
        if (strCaseCmp(name, i->substr(0, name.size())) == 0)
            return strLTrim(i->substr(name.size()));
    return std::string();
}


#ifdef DOWNLOADER
#include <iostream>
int main()
{
    std::string s;
    Url u;
    while (getline(std::cin, s)) {
        if (!u.parse(s)) {
            std::cout << "Ungueltig.\n";
        } else {
            std::cout << "Host = '" << u.getHost() << "'\n"
                      << "Path = '" << u.getPath() << "'\n"
                      << "Proto = '" << u.getProtocol() << "'\n"
                      << "Port = " << u.getPort() << "\n"
                      << "==> '" << u.toString() << "'\n";
            if (u.getProtocol() != "http")
                std::cout << "==> can only do http\n";
            else if (!Http().downloadFile(u, std::cout))
                std::cout << "==> download failed\n";
            else
                std::cout << "==> download ok\n";

            std::cout << "    basename is '" << u.getBaseName() << "'\n";

            // Parameter
            class E : public KVEnumerator {
             public:
                void handleItem(std::string a, std::string, std::string b)
                    { std::cout << a << " = " << b << "\n"; }
            } e;
            u.trimParameters(e);
        }
    }
    return 0;
}
#endif

#ifdef UPLOADER
#include <iostream>
#include <fstream>
int main(int argc, char** argv)
{
    if (argc != 4) {
        std::cout << "usage: uploader URL NAME FILE\n";
        return 1;
    }
    std::string s;
    Url u;
    if (!u.parse(argv[1])) {
        std::cout << "invalid url\n";
        return 1;
    }
    MimeBuilder builder("multipart/form-data");
    MimeParameterEncoder mpe(builder);
    u.trimParameters(mpe);

    // Now add the file
    std::ifstream in(argv[3], std::ios::binary);
    if (in) {
        builder.addBoundary();
        builder.addHeader("Content-Disposition",
                          std::string("form-data; name=\"") + argv[2] + "\"; filename=\"" + argv[3] + "\"");
        builder.addHeader("Content-Type", "application/octet-stream");
        builder.addHeader("Content-Transfer-Encoding", "base64");
        char buf[4096];
        int count;
        while ((count = in.rdbuf()->sgetn(buf, sizeof(buf))) > 0) {
            builder.addBase64(buf, count);
        }
    } else {
        std::cerr << "Can't open " << *argv << "\n";
        return 1;
    }
    builder.addBoundary();
    builder.finish();

    // And send it
    std::cout << "Now posting to " << u.toString() << "...\n";
    if (Http::post(u, builder, std::cout))
        std::cout << "OK\n";
    else
        std::cout << "Failure\n";
    return 0;
}
#endif
