/**
  *  \file mimebuilder.h
  *  \brief Building MIME entities
  *  \author Stefan Reuther
  */
#ifndef MIMEBUILDER_H_INCLUDED
#define MIMEBUILDER_H_INCLUDED

#include <list>
#include <string>
#include <cstddef>
#include <ostream>
#include "enumerator.h"

/** Simple MIME Multipart Entity Builder. This gathers up a MIME
    message line by line, and fills in the boundaries.

    Usage:
    - construct an object <tt>MimeBuilder b("content-type");</tt>
      giving it the multipart content type as parameter
    - use addHeader(...) to add additional headers
    - use addBoundary() followed by multiple addHeader(), followed
      by addLine()/addBase64()/addLineQP() to add parts
    - use addBoundary() to add the final boundary
    - use finish() to finish the object
    - possibly use removeInitialHeaders()
    - use computeSize() and write() to get the data

    Content-Transfer-Encoding is not managed by MimeBuilder, you must
    write a Content-Transfer-Encoding line that matches your addXxx()
    calls yourself. */
class MimeBuilder {
    /** Type of an element. */
    enum ElementType {
        TextElement,              /**< Raw text. */
        ContentTypeElement,       /**< Content-type line. */
        BoundaryElement           /**< Boundary. */
    };
    /** Element (line) of our message. */
    struct Element {
        std::string m_Text;
        ElementType m_Type;

        Element(const std::string& text, ElementType type)
            : m_Text(text), m_Type(type) { }
        
    };
    /** Content of our message. */
    std::list<Element> m_Content;

    /** Content-Type to use. */
    std::string m_ContentType;

    /** Accumulator for building base64 parts. */
    std::string m_Base64Accumulator;

    /** State. */
    enum State {
        StateHeader,
        StateBody
    };
    State m_State;

    void flushHeaderAndStartBody();
    void flushBase64();
 public:
    MimeBuilder(std::string ctype);

    ~MimeBuilder();

    void addHeader(std::string name, std::string value);

    void addLine(std::string line);

    void addBase64(const char* data, std::size_t size);

    void addLineQP(std::string line);

    void addBoundary();

    void finish();

    std::size_t computeSize() const;

    void write(std::ostream& out);

    std::string removeInitialHeaders();
};

/** Simple parameter encoder. Can be used with Url::trimParameters()
    or standalone to add small parameter blocks to a MimeBuilder. */
class MimeParameterEncoder : public KVEnumerator {
    MimeBuilder& m_Builder;
 public:
    MimeParameterEncoder(MimeBuilder& builder)
        : m_Builder(builder)
        { }

    void handleItem(std::string key, std::string description, std::string value);
};

#endif
