/**
  *  \file smtp.cpp
  *  \brief Sending Mail
  *  \author Stefan Reuther
  *
  *  References:
  *
  *  RFC  821  SMTP
  *  RFC  822  Mail message format
  *  RFC 1321  MD5
  *  RFC 1869  ESMTP
  *              - specifies the EHLO command, and responses of the form
  *                         250-servername ESMTP hi there
  *                         250-FOO
  *                         250-BAR
  *                         250 HUHU
  *  RFC 2104  HMAC-MD5
  *  RFC 2195  CRAM-MD5
  *  RFC 2554  SMTP-Auth
  *  RFC 2595  AUTH PLAIN


 {smtp to mail.gmx.net} Verbindung aufbauen...
 TClientSocketSMTP.OnLookup from Hamster to mail.gmx.net:smtp use mail.gmx.net:25
 TClientSocketSMTP.OnConnecting from Hamster to mail.gmx.net:smtp use mail.gmx.net:25
 TClientSocketSMTP.OnConnect from Hamster to mail.gmx.net:smtp use mail.gmx.net:25
 TClientSocketSMTP > EHLO royale
 TClientSocketSMTP < 250 STARTTLS
 TClientSocketSMTP > AUTH CRAM-MD5
 TClientSocketSMTP < 334 PDYwNDguMTE2MzMzMjIwNkBtcDAxOT4=
 TClientSocketSMTP > [...authorization...]
 TClientSocketSMTP < 235 2.7.0 Go ahead {mp019}
 {smtp to mail.gmx.net} Verschicke E-Mail "53363.msg"
 [mail.gmx.net, SMTP] SendMail 53363.msg
 TClientSocketSMTP > MAIL FROM:<streu@gmx.de>
 TClientSocketSMTP < 250 2.1.0 ok {mp019}
 TClientSocketSMTP > RCPT TO:<zap>
 TClientSocketSMTP < 250 2.1.5 ok {mp019}
 TClientSocketSMTP > DATA
 TClientSocketSMTP < 354 mail.gmx.net Go ahead {mp019}
 TClientSocketSMTP > .
 TClientSocketSMTP < 250 2.6.0 Message accepted {mp019}
 E-Mail 53363.msg wurde verschickt.
 {smtp to mail.gmx.net} Verbindung abbauen...
 TClientSocketSMTP > QUIT
 TClientSocketSMTP < 221 2.0.0 GMX Mailservices {mp019}

  
  */

#include <cstring>
#include <cstdlib>
#include "net/smtp.h"
#include "net/socketstream.h"
#include "socket.h"
#include "nls.h"

static std::string smtp_server;
static int smtp_port = 25;
static std::string smtp_from_addr;

void
Smtp::setServer(std::string server, int port)
{
    smtp_server = server;
    smtp_port = port >= 0 ? port : 25;
}

void
Smtp::setFromAddress(std::string addr)
{
    smtp_from_addr = addr;
}

std::string
Smtp::getServer()
{
    return smtp_server;
}

int
Smtp::getPort()
{
    return smtp_port;
}

std::string
Smtp::getFromAddress()
{
    return smtp_from_addr;
}

bool
Smtp::isConfigured()
{
    return smtp_server.size() && smtp_from_addr.size() && smtp_port > 0;
}

bool
Smtp::readReply(int expect)
{
    std::string line;
    while (readLine(line)) {
        if (line.size() >= 4 && line[3] != '-') {
            int code = std::atoi(line.c_str()) / 100;
            if (code != expect) {
                setMessage(_("Unexpected response from server"), line);
                return false;
            } else {
                return true;
            }
        }
        line.clear();
    }
    return -1;
}

/** Send mail via server.
    \param from     [in] Our address for SMTP envelope
    \param to       [in] Receiver address for SMTP envelope
    \param content  [in] Mail data
    \return true if mail successfully submitted at server */
bool
Smtp::sendMail(std::string from, std::string to, MimeBuilder& content)
{
    // figure out our host name
    char hostname[256];
    if (gethostname(hostname, sizeof(hostname)) == 0) {
        hostname[sizeof(hostname)-1] = 0;
    } else {
        std::strcpy(hostname, "localhost");
    }

    if (!connect(smtp_server, smtp_port))
        return false;

    // receive greeting
    if (!readReply(2)) {
        close();
        return false;
    }

    // hello
    std::string command = "HELO ";
    command += hostname;
    command += "\015\012";
    if (!write(command) || !readReply(2)) {
        close();
        return false;
    }

    // mail from
    command = "MAIL FROM:<" + from + ">\015\012";
    if (!write(command) || !readReply(2)) {
        close();
        return false;
    }

    // rcpt to
    command = "RCPT TO:<" + to + ">\015\012";
    if (!write(command) || !readReply(2)) {
        close();
        return false;
    }

    // data
    command = "DATA\015\012";
    if (!write(command) || !readReply(3)) {
        close();
        return false;
    }

    // send the mail
    {
        SocketStream ss(m_Socket);
        content.write(ss);
    }
    command = ".\015\012";
    if (!write(command) || !readReply(2)) {
        close();
        return false;
    }

    // close
    command = "QUIT\015\012";
    if (!write(command) || !readReply(2)) {
        close();
        return false;
    }

    close();
    return true;
}

#ifdef MAILER
#include <fstream>
#include <iostream>
int main(int argc, char** argv)
{
    if (argc != 5) {
        puts("mailer SERVER FROM TO-URL FILE");
        return 1;
    }

    Smtp::setServer(argv[1], -1);
    
    MimeBuilder mb("multipart/mixed");

    /* Mail headers */
    mb.addHeader("From", argv[2]);
    mb.addHeader("To", argv[3]);
    mb.addHeader("Subject", "Turn file for <game>");
    mb.addHeader("X-Mailer", "VGAPIntegrator/0.5");
    mb.addHeader("MIME-Version", "1.0");

    /* Text part */
    mb.addBoundary();
    mb.addHeader("Content-Type", "text/plain; charset=iso-8859-1");
    mb.addHeader("Content-Disposition", "inline");
    mb.addHeader("Content-Transfer-Encoding", "quoted-printable");
    mb.addLineQP("Please find attached the turn file for <game> player <race>.");
    mb.addLineQP("This mail was sent with VGAPIntegrator/0.5");
    mb.addLineQP("");

    /* File part */
    mb.addBoundary();
    mb.addHeader("Content-Type", "application/octet-stream");
    mb.addHeader("Content-Disposition", "attachment; filename=\"" + std::string(argv[4]) + "\"");
    mb.addHeader("Content-Transfer-Encoding", "base64");

    {
        std::ifstream in(argv[4], std::ios::binary);
        if (in) {
            char buf[4096];
            int count;
            while ((count = in.rdbuf()->sgetn(buf, sizeof(buf))) > 0) {
                mb.addBase64(buf, count);
            }
        } else {
            puts("can't open file");
            return 1;
        }
    }
    mb.addBoundary();
    mb.finish();
    std::cout << "Now mailing...\n";
    if (Smtp().sendMail(argv[2], argv[3], mb)) {
        puts("mail sent ok");
    } else {
        puts("error");
        return 1;
    }
    return 0;
}
#endif
