/**
  *  \file util.cpp
  */

#include <string>
#include <stdexcept>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include "util.h"

const char* const bool_table[] = { "no", "yes", 0 };

/** Trim whitespace. Removes whitespace from both sides.
    \param s [in] Input
    \return s, with leading and trailing whitespace removed. */
std::string
strTrim(std::string s)
{
    return strLTrim(strRTrim(s));
}

/** Trim leading (left) whitespace.
    \param s [in] Input
    \return s, with leading whitespace (see #isspace) removed. */
std::string
strLTrim(std::string s)
{
    std::string::size_type n = 0;
    while (n < s.length() && isspace(s[n]))
        ++n;
    return s.substr(n);
}

/** Trim trailing (right) whitespace.
    \param s [in] Input
    \return s, with trailing whitespace (see #isspace) removed. */
std::string
strRTrim(std::string s)
{
    std::string::size_type n = s.length();
    while (n > 0 && isspace(s[n-1]))
        --n;
    return s.substr(0, n);
}

/** Convert integer to string. Returns the number string. */
std::string
itoa(int i)
{
    char buffer[100];
    std::sprintf(buffer, "%d", i);
    return std::string(buffer);
}

/** Convert string to integer.
    \param s [in] String to convert
    \param i [out] Result
    \return true if conversion succeded, otherwise false */
bool
atoi(std::string s, int& i)
{
    const char* p = s.c_str();
    char* endp;
    long l = strtol(p, &endp, 0);
    if (*p != 0 && *endp == 0) {
        i = l;
        return true;
    } else {
        return false;
    }
}

/** Convert string to lower case. */
std::string
strLower(std::string s)
{
    for (std::string::size_type i = 0; i < s.size(); ++i)
        s[i] = tolower((unsigned char) s[i]);
    return s;
}

/** Eat word from a string. Removes the first space-delimited word
    from s, and returns it.
    \param s [in/out] String containing words
    \return extracted word */
std::string
eatWord(std::string& s)
{
    std::string::size_type beg = 0;
    while (beg < s.size() && isspace((unsigned char) s[beg]))
        ++beg;

    std::string::size_type end = beg;
    while (end < s.size() && !isspace((unsigned char) s[end]))
        ++end;

    std::string::size_type rest = end;
    while (rest < s.size() && isspace((unsigned char) s[rest]))
        ++rest;

    std::string result(s, beg, end-beg);
    s.erase(0, rest);
    return result;
}

/** Compare two strings, ignoring case. */
int
strCaseCmp(const std::string a, const std::string b)
{
    using namespace std;
    std::string::size_type max = a.size() > b.size() ? b.size() : a.size();
    for (std::string::size_type i = 0; i < max; ++i) {
        unsigned char ac = toupper((unsigned char) a[i]);
        unsigned char bc = toupper((unsigned char) b[i]);
        if (ac != bc)
            return ac < bc ? -1 : +1;
    }
    if (a.size() != b.size())
        return a.size() < b.size() ? -1 : +1;
    return 0;
}

/** Extract parameter value. */
std::string
getParameter(std::string line, std::string parameter)
{
    /* I'm not 100% sure about this one's standard conformity, as I
       have not found a definitive syntax definition of a
       parameterized header line. This handles what we generate, and
       what planetsserver does. */
    bool inquote = false;
    bool inbs    = false;
    bool invalue = false;
    std::string pair[2];

    for (std::string::size_type n = 0; n < line.size(); ++n) {
        char c = line[n];
        if (inbs) {
            pair[invalue] += c;
            inbs = false;
        } else if (inquote) {
            if (c == '"')
                inquote = false;
            else if (c == '\\')
                inbs = true;
            else
                pair[invalue] += c;
        } else {
            if (c == '"') {
                inquote = true;
            } else if (c == ';') {
                if (strCaseCmp(strTrim(pair[0]), parameter) == 0)
                    return strTrim(pair[1]);
                invalue = false;
                pair[0].clear();
                pair[1].clear();
            } else if (c == '=') {
                invalue = true;
            } else {
                pair[invalue] += c;
            }
        }
    }
    if (strCaseCmp(strTrim(pair[0]), parameter) == 0)
        return strTrim(pair[1]);
    return std::string();
}

/** Failure report using exception. */
void
fail(std::string what)
{
    throw std::runtime_error(what);
}

/** Find index of a word in a string list.
    \param text   [in] Text to look for
    \param table  [in] List of null-terminated strings
    \param result [out] Index such that table[result] == text on success
    \return true on success */
bool
stringToIntFromList(std::string text, const char*const* table, int& result)
{
    int n = 0;
    while (*table) {
        if (text == *table) {
            result = n;
            return true;
        }
        ++table, ++n;
    }
    return false;
}
