/**
  *  \file execute.h
  *  \brief Executing stuff
  */
#include <string>
#include <fstream>
#include <sstream>
#include "execute.h"
#include "game.h"
#include "selection.h"
#include "race.h"
#include "action.h"
#include "net/smtp.h"
#include "net/http.h"
#include "net/mailtemplate.h"
#include "nls.h"
#include "net/url.h"
#include "osdep.h"

/* Turn uploading */

class TurnUploader : public ValueEnumerator<Race> {
    Game& m_Game;
    HWND m_Wnd;
 public:
    TurnUploader(Game& g, HWND wnd)
        : m_Game(g), m_Wnd(wnd)
        { }

    void handleItem(Race&);

    void doUpload(Url& u, Race& r);

    void doMail(const MailTemplate& t, Race& r);
};

class ResultDownloader : public ValueEnumerator<Race> {
    Game& m_Game;
    Action& m_Action;
    HWND m_Window;
 public:
    ResultDownloader(Game& g, Action& a, HWND w)
        : m_Game(g), m_Action(a), m_Window(w)
        { }

    void handleItem(Race&);
};

static void
stripTags(std::string& s)
{
    std::string::size_type n, m;
    while ((n = s.find('<')) != s.npos && (m = s.find('>', n)) != s.npos)
        s.erase(n, m-n);
}

static void
stripBlanks(std::string& s)
{
    std::string::size_type n = 0;
    while (n < s.size()) {
        if (s[n] == ' ' || s[n] == '\n' || s[n] == '\t' || s[n] == '\r') {
            bool had_nl = false;
            std::string::size_type q = n;
            while (q < s.size() && (s[q] == ' ' || s[q] == '\n' || s[q] == '\t' || s[q] == '\r')) {
                if (s[q] == '\n')
                    had_nl = true;
                ++q;
            }
            s.erase(n, q-n);
            if (had_nl) {
                s[n] = '\n';
                s.insert(n++, "\r");
            } else {
                s[n] = ' ';
            }
        }
        ++n;
    }
}

void
TurnUploader::doUpload(Url& u, Race& r)
{
    // upload race's turn file
    std::string basename = "player" + itoa(r.getRaceNumber()) + ".trn";
    std::string tname = makeFileName(m_Game.getLocalDirectory(), basename);
    std::ifstream trn(tname.c_str(), std::ifstream::binary);
    if (!trn) {
        MessageBox(m_Wnd, (_("The file '") + basename + _("' does not exist. Run maketurn first.")).c_str(),
                   _("Send turn files"),
                   MB_OK);
        return;
    }

    // build packet to upload
    MimeBuilder builder("multipart/form-data");
    MimeParameterEncoder mpe(builder);
    u.trimParameters(mpe);

    // Now add the file
    builder.addBoundary();
    builder.addHeader("Content-Disposition",
                      std::string("form-data; name=\"") + r.getUploadId() + "\"; filename=\"" + basename + "\"");
    builder.addHeader("Content-Type", "application/octet-stream");
    builder.addHeader("Content-Transfer-Encoding", "base64");
    char buf[4096];
    int count;
    while ((count = trn.rdbuf()->sgetn(buf, sizeof(buf))) > 0) {
        builder.addBase64(buf, count);
    }
    builder.addBoundary();
    builder.finish();

    // Upload
    Http req;
    std::stringstream accum;
    if (req.post(u, builder, accum)) {
        if (!req.isSuccessfulReply()) {
            MessageBox(m_Wnd, (_("Turn upload failed: server said '") + req.getReplyLine() + "'").c_str(),
                       _("Send turn files"),
                       MB_ICONSTOP | MB_OK);
        } else {
            // MessageBox(m_Wnd, (_("Turn file '") + basename + _("' successfully uploaded.")).c_str(),
            //            _("Send turn files"),
            //            MB_ICONSTOP | MB_OK);
            std::string s = accum.str();
            stripTags(s);
            stripBlanks(s);
            MessageBox(m_Wnd, (_("Result of turn file upload:\r\n") + s).c_str(),
                       _("Send turn files"),
                       MB_ICONSTOP | MB_OK);
        }
    } else {
        MessageBox(m_Wnd, (_("Turn upload failed: ") + req.getErrorMessage()
                           + _("\r\nNetwork message: ") + req.getServerErrorMessage()).c_str(),
                   _("Send turn files"),
                   MB_ICONSTOP | MB_OK);
    }
}

void
TurnUploader::doMail(const MailTemplate& t, Race& r)
{
    if (!Smtp::isConfigured()) {
        MessageBox(m_Wnd, _("Mail not configured. Use 'File->Network Configuration' to configure "
                            "your mail address and server to be able to mail turn files."),
                   _("Send turn files"),
                   MB_ICONSTOP | MB_OK);
        return;
    }

    // mail race's turn file
    std::string basename = "player" + itoa(r.getRaceNumber()) + ".trn";
    std::string tname = makeFileName(m_Game.getLocalDirectory(), basename);
    std::ifstream trn(tname.c_str(), std::ifstream::binary);
    if (!trn) {
        MessageBox(m_Wnd, (_("The file '") + basename + _("' does not exist. Run maketurn first.")).c_str(),
                   _("Send turn files"),
                   MB_OK);
        return;
    }

    // build packet to upload
    MimeBuilder mb("multipart/mixed");
    mb.addHeader("From", Smtp::getFromAddress());
    mb.addHeader("To", t.getAddress());
    if (t.getSubject().size())
        mb.addHeader("Subject", t.getSubject());
    else
        mb.addHeader("Subject", "Turn file for " + m_Game.getName());
    mb.addHeader("X-Mailer", "VGAPIntegrator/0.5");
    mb.addHeader("MIME-Version", "1.0");

    /* Text part */
    mb.addBoundary();
    mb.addHeader("Content-Type", "text/plain; charset=iso-8859-1");
    mb.addHeader("Content-Disposition", "inline");
    mb.addHeader("Content-Transfer-Encoding", "quoted-printable");
    if (t.getBody().size()) {
        mb.addLineQP(t.getBody());
    } else {
        mb.addLineQP("Please find attached the turn file for '" + m_Game.getName() + "' player " +
                     itoa(r.getRaceNumber()) + ".");
    }
    mb.addLineQP("");

    /* File part */
    mb.addBoundary();
    mb.addHeader("Content-Type", "application/octet-stream");
    mb.addHeader("Content-Disposition", "attachment; filename=\"" + basename + "\"");
    mb.addHeader("Content-Transfer-Encoding", "base64");
    char buf[4096];
    int count;
    while ((count = trn.rdbuf()->sgetn(buf, sizeof(buf))) > 0) {
        mb.addBase64(buf, count);
    }
    mb.addBoundary();
    mb.finish();

    Smtp mailer;
    if (mailer.sendMail(Smtp::getFromAddress(), t.getAddress(), mb)) {
        MessageBox(m_Wnd, (_("Turn file '") + basename + _("' successfully mailed.")).c_str(),
                   _("Send turn file"),
                   MB_OK);
    } else {
        MessageBox(m_Wnd, (_("Mailing turn failed: ") + mailer.getErrorMessage()
                           + _("\r\nNetwork message: ") + mailer.getServerErrorMessage()).c_str(),
                   _("Send turn files"),
                   MB_ICONSTOP | MB_OK);
    }
}

void
TurnUploader::handleItem(Race& r)
{
    // figure out URL
    {
        Url u;
        if (u.parse(r.getUploadUrl())) {
            if (u.getProtocol() != "http" && u.getProtocol() != "http-post") {
                MessageBox(m_Wnd, _("The specified upload URL is not currently supported."),
                           _("Configuration error"),
                           MB_ICONSTOP | MB_OK);
                return;
            }
            doUpload(u, r);
            return;
        }
    }

    // or do we want to mail?
    {
        MailTemplate t;
        if (t.parse(r.getUploadUrl())) {
            doMail(t, r);
            return;
        }
    }

    MessageBox(m_Wnd, (_("The specified upload URL '") + r.getUploadUrl() + _("' is not currently supported.")).c_str(),
               _("Configuration error"),
               MB_ICONSTOP | MB_OK);
}

static void
doDownloadFile(HWND window, std::string url, Game& g)
{
    Url u;
    if (!u.parse(url) || (u.getProtocol() != "http" && u.getProtocol() != "http-post")) {
        MessageBox(window, (_("The specified URL '") + url + "' is not currently supported.").c_str(),
                   _("Configuration error"),
                   MB_ICONSTOP | MB_OK);
        return;
    }

    std::string tmpname = makeFileName(g.getLocalDirectory(), "__vgapi.tmp");
    std::ofstream out(tmpname.c_str(), std::ofstream::binary);
    if (!out) {
        MessageBox(window, _("Unable to create temporary file."),
                   _("File download"),
                   MB_ICONSTOP | MB_OK);
        return;
    }

    Http downloader;
    bool result;
    if (u.getProtocol() == "http") {
        // GET
        result = downloader.downloadFile(u, out);
    } else {
        // POST
        MimeBuilder mb("multipart/form-data");
        MimeBuilder builder("multipart/form-data");
        MimeParameterEncoder mpe(mb);
        u.trimParameters(mpe);
        mb.addBoundary();
        mb.finish();
        result = downloader.post(u, mb, out);
    }
    out.close();

    if (!result) {
        MessageBox(window, (_("File download failed: ") + downloader.getErrorMessage()
                            + _("\r\nNetwork message: ") + downloader.getServerErrorMessage()).c_str(),
                   _("File download"),
                   MB_ICONSTOP | MB_OK);
        std::remove(tmpname.c_str());
    } else if (!downloader.isSuccessfulReply()) {
        MessageBox(window, (_("File download failed: server said '") + downloader.getReplyLine() + "'").c_str(),
                   _("File download"),
                   MB_ICONSTOP | MB_OK);
        std::remove(tmpname.c_str());
    } else {
        // Figure out the name of the file
        std::string outname;
        outname = getParameter(downloader.getHeader("Content-Disposition"), "filename");
        if (outname.size() == 0)
            outname = getParameter(downloader.getHeader("Content-Type"), "filename");
        if (outname.size() == 0)
            outname = u.getBaseName();

        outname = getBaseName(outname);

        // uh-oh
        std::remove(outname.c_str());
        std::rename(tmpname.c_str(), makeFileName(g.getLocalDirectory(), outname).c_str());

        MessageBox(window, (_("File '") + outname + _("' successfully downloaded.")).c_str(),
                   _("File download"),
                   MB_OK);
    }
}

void
ResultDownloader::handleItem(Race& r)
{
    Selection sel;
    sel.selectGame(&m_Game);
    sel.selectRace(&r);
    sel.selectAction(&m_Action);

    std::string url = sel.expandString(r.getDownloadUrl());
    if (url.size() == 0) {
        MessageBox(m_Window, (_("The result download URL for player ") + itoa(r.getRaceNumber())
                              + _(" was not specified.")).c_str(),
                   _("Configuration error"),
                   MB_ICONSTOP | MB_OK);
    } else {
        doDownloadFile(m_Window, url, m_Game);
    }
}


/* Main Entry Point */

void
runAction(HWND window, Action& a, Selection& sel)
{
    sel.selectAction(&a);
    std::string s = sel.expandString(a.getCommandLine());
    if (s.size() == 0) {
        MessageBox(window, _("This command is currently not possible."), a.getDescription().c_str(),
                   MB_ICONSTOP | MB_OK);
    } else {
        std::string verb = eatWord(s);
        if (verb == "command") {
            while (s.size() && s[0] == '-')
                eatWord(s);
            MessageBox(window, (_("Would now run '") + s + "'\r\n").c_str(), a.getDescription().c_str(),
                       MB_ICONSTOP | MB_OK);
        } else if (verb == "browser") {
            if (s.size()) {
                ShellExecute(window, "open", s.c_str(), 0,
                             sel.getSelectedGame() ? sel.getSelectedGame()->getLocalDirectory().c_str() : ".",
                             0);
            } else {
                MessageBox(window, _("This action would open a web site, but the URL was not specified."),
                           _("Configuration error"),
                           MB_ICONSTOP | MB_OK);
            }
        } else if (verb == "builtin-download-rst") {
            if (sel.getSelectedGame()) {
                ResultDownloader rd(*sel.getSelectedGame(), a, window);
                sel.getSelectedGame()->enumerateRaces(rd);
            }
        } else if (verb == "builtin-download-spec") {
            if (!s.size()) {
                MessageBox(window, _("This action would download a file, but the URL was not specified."),
                           _("Configuration error"),
                           MB_ICONSTOP | MB_OK);
            } else if (sel.getSelectedGame()) {
                doDownloadFile(window, s, *sel.getSelectedGame());
            }
        } else if (verb == "builtin-upload-trn") {
            if (sel.getSelectedGame()) {
                TurnUploader tu(*sel.getSelectedGame(), window);
                sel.getSelectedGame()->enumerateRaces(tu);
            }
        } else {
            MessageBox(window, (_("The command '") + verb + _("' specified by this action is not supported.")).c_str(),
                       _("Configuration error"), MB_ICONSTOP | MB_OK);
        }
    }
}
